# Third party
from django.utils.translation import gettext as _
from django.utils.translation import pgettext_lazy

# Local application / specific library imports
from ..checks import custom_list, utils


def importance():
    """Scripts with higher importance will be executed first.

    Returns:
        int -- Importance of the script.
    """
    return 1


def run(site):
    """Verifies that only one h1 tag is present, and that it contains at least one keyword.

    Arguments:
        site {Site} -- Structure containing a good amount of resources from the targeted webpage.
    """
    too_much_h1 = custom_list.CustomList(
        name=_("Too much h1 tags"),
        settings=_("exactly 1"),
        description=_(
            "Google is not really concerned about the number of h1 tags on your page, but Bing clearly indicates in its guidelines for webmasters to use only one h1 tag per page."
        ),
    )

    right_number_h1 = custom_list.CustomList(
        name=_("H1 tag found"),
        settings=_("exactly 1"),
        description=too_much_h1.description,
    )

    not_enough_h1 = custom_list.CustomList(
        name=_("No h1 tag"),
        settings=_("exactly 1"),
        description=too_much_h1.description,
    )

    no_keywords = custom_list.CustomList(
        name=_("No keyword in h1"),
        settings=_("at least one"),
        description=_(
            "The h1 tag represent the main title of your page, and you may populate it with appropriate content in order to ensure that users (and search engines!) will understand correctly your page."
        ),
    )

    enough_keywords = custom_list.CustomList(
        name=_("Keyword found in h1"),
        settings=_("at least one"),
        description=no_keywords.description,
    )

    h1_all = site.soup.find_all("h1")

    if len(h1_all) > 1:
        too_much_h1.found = len(h1_all)
        too_much_h1.searched_in = [utils.get_heading_text(t) for t in h1_all]
        site.problems.append(too_much_h1)

    elif not h1_all:
        not_enough_h1.found = pgettext_lazy("masculin", "none")
        site.problems.append(not_enough_h1)

    else:
        right_number_h1.found = len(h1_all)
        right_number_h1.searched_in = [utils.get_heading_text(t) for t in h1_all]
        site.success.append(right_number_h1)

    enough_keywords.found = ""
    h1_text_kw = []
    occurrence = []
    for h1 in h1_all:
        text = utils.get_heading_text(h1).lower()

        highlighted, occ, found = utils.highlight_keywords_in_text(
            text, site.keywords, normalize_apostrophes_flag=True
        )
        occurrence.extend(occ)
        h1_text_kw.append(highlighted)
        if found:
            enough_keywords.found += (", " if enough_keywords.found else "") + found

    # if no keyword is found in h1
    if not any(i > 0 for i in occurrence):
        no_keywords.found = pgettext_lazy("masculin", "none")
        no_keywords.searched_in = [t.text for t in h1_all]
        site.problems.append(no_keywords)
    else:
        enough_keywords.searched_in = h1_text_kw
        site.success.append(enough_keywords)



import importlib
import json
import re
from urllib.parse import quote, urlparse

from bs4 import BeautifulSoup
from django.contrib.auth.mixins import PermissionRequiredMixin
from django.shortcuts import render
from django.test import Client
from django.utils.translation import gettext as _
from django.utils.translation import ngettext
from django.views import generic

from .checks import site
from .checks_list import launch_checks
from .conf import settings
from .models import Keyword, Page
from .utils.cms_url_resolution import (
    get_cms_url_for_check_seo,
    get_preview_url_for_check_seo,
)
from .utils.keywords_discovery import get_cms_page_id_language_for_path


def _quote_path(path):
    """Quote path for use in query string (encode slashes so one param is parsed correctly)."""
    return quote(path or "", safe="")


def _parse_keywords_text(text):
    """Parse keywords from text (comma or newline separated), return list of stripped non-empty strings."""
    if not text or not text.strip():
        return []
    parts = re.split(r"[\n,]+", text)
    return [p.strip() for p in parts if p.strip()]


def _normalize_path(path):
    if not path:
        return ""
    path = (path or "").strip()
    if path and not path.startswith("/"):
        path = "/" + path
    return path or "/"


def _get_keywords_for_path(path):
    """Return list of keyword names for the given page path, or []."""
    path = _normalize_path(path)
    if not path or path == "/":
        return []
    try:
        page = Page.objects.get(path=path)
        return [kw.name for kw in page.keywords.all()]
    except Page.DoesNotExist:
        return []


def _get_keywords_for_cms_page(cms_page_id, language):
    """Return list of keyword names for the given (cms_page_id, language), or []."""
    if cms_page_id is None:
        return []
    page = Page.objects.filter(cms_page_id=cms_page_id, language=language or "").first()
    if not page:
        return []
    return [kw.name for kw in page.keywords.all()]


def _get_keywords_for_page_identifier(path=None, cms_page_id=None, language=None):
    """Return keyword names for the page identified by path or (cms_page_id, language)."""
    if cms_page_id is not None:
        return _get_keywords_for_cms_page(cms_page_id, language)
    return _get_keywords_for_path(path or "")


def _get_or_create_page_for_edit(path=None, cms_page_id=None, language=None):
    """
    Get or create Page by path or (cms_page_id, language). Return (page, path_display, use_cms_id).
    """
    if cms_page_id is not None:
        page, _ = Page.objects.get_or_create(
            cms_page_id=cms_page_id,
            language=language or "",
            defaults={"path": None},
        )
        path_display = page.path or f"/cms-page-{cms_page_id}/"
        return (page, path_display, True)
    path = _normalize_path(path)
    if not path:
        path = "/"
    page, _ = Page.objects.get_or_create(path=path, defaults={})
    return (page, path, False)


def _get_all_keyword_names():
    """Return sorted list of all keyword names (for datalist autocomplete)."""
    return list(
        Keyword.objects.values_list("name", flat=True).distinct().order_by("name")
    )


def _get_keywords_discovery_function():
    """Resolve the keywords discovery function from settings (import path)."""
    method = settings.DJANGO_CHECK_SEO_KEYWORDS_DISCOVERY_METHOD
    module_path, func_name = method.rsplit(".", 1)
    module = importlib.import_module(module_path)
    return getattr(module, func_name)


class IndexView(PermissionRequiredMixin, generic.base.TemplateView):
    template_name = "django_check_seo/default.html"
    permission_required = "django_check_seo.use_django_check_seo"

    def get_context_data(self, *args, **kwargs):
        context = super(generic.base.TemplateView, self).get_context_data(
            *args, **kwargs
        )

        page = self.request.GET.get("page")
        preview_url = get_preview_url_for_check_seo(page)
        if preview_url:
            url_to_fetch = preview_url
            client = Client()
            client.force_login(self.request.user)
            response = client.get(url_to_fetch, follow=True)
            analysis_draft = True
        else:
            url_to_fetch = get_cms_url_for_check_seo(page)
            if url_to_fetch is None:
                url_to_fetch = page
            client = Client()
            response = client.get(url_to_fetch, follow=True)
            analysis_draft = False

        soup = BeautifulSoup(response.content, features="lxml")

        # populate class with data (use url_to_fetch so full_url is the page we actually checked)
        page_stats = site.Site(soup, url_to_fetch)

        # populate keywords using configured discovery method (meta or model)
        discovery_func = _get_keywords_discovery_function()
        discovery_func(page_stats)

        # magic happens here!
        launch_checks.launch_checks(page_stats)

        # end of magic, get collected problems/warnings/success and put them inside the context
        (context["problems"], context["warnings"], context["success"]) = (
            page_stats.problems,
            page_stats.warnings,
            page_stats.success,
        )

        context["analysis_draft"] = analysis_draft
        context["settings"] = json.dumps(settings.DJANGO_CHECK_SEO_SETTINGS, indent=4)
        context["html"] = page_stats.content
        context["text"] = page_stats.content_text
        context["keywords"] = page_stats.keywords
        discovery_method = settings.DJANGO_CHECK_SEO_KEYWORDS_DISCOVERY_METHOD
        context["keywords_editable"] = settings.DJANGO_CHECK_SEO_KEYWORDS_EDITABLE and (
            "model_keywords" in discovery_method
            or "django_cms_model_keywords" in discovery_method
        )
        # For form: identify page by path and/or (cms_page_id, language) when using django_cms_model_keywords
        path_for_display = url_to_fetch
        path_for_resolution = (
            urlparse(url_to_fetch or "/").path if url_to_fetch else "/"
        )
        if not path_for_resolution or path_for_resolution == "/":
            path_for_resolution = "/"
        cms_page_id, cms_language = get_cms_page_id_language_for_path(
            path_for_resolution
        )
        use_cms_page_id = (
            context["keywords_editable"]
            and "django_cms_model_keywords" in discovery_method
            and cms_page_id is not None
        )
        context["current_page_path"] = path_for_display
        context["use_cms_page_id"] = use_cms_page_id
        context["current_page_cms_id"] = cms_page_id if use_cms_page_id else None
        context["current_page_language"] = cms_language if use_cms_page_id else None
        if context["keywords_editable"]:
            base = self.request.path.rstrip("/")
            context["keywords_edit_url"] = base + "/keywords-edit/"
            if use_cms_page_id:
                context["keywords_cancel_url"] = (
                    base
                    + "/keywords-edit/?cms_page_id="
                    + str(cms_page_id)
                    + "&language="
                    + quote(cms_language or "", safe="")
                    + "&show_form=0"
                )
                context["checks_refresh_url"] = (
                    self.request.path
                    + "?page="
                    + _quote_path(path_for_display)
                    + "&fragment=checks"
                )
            else:
                context["keywords_cancel_url"] = (
                    base
                    + "/keywords-edit/?path="
                    + _quote_path(path_for_display)
                    + "&show_form=0"
                )
                context["checks_refresh_url"] = (
                    self.request.path
                    + "?page="
                    + _quote_path(path_for_display)
                    + "&fragment=checks"
                )
            context["all_keywords"] = _get_all_keyword_names()

        nb_problems = len(context["problems"])
        nb_warnings = len(context["warnings"])

        # define some fancy-looking text here because it is wayyy to dirty with all the {% translate%} in template
        if nb_problems == 0 and nb_warnings == 0:
            context["nb_problems_warnings"] = _(
                '<strong class="green-bg">No problem</strong> was found on the page!'
            )
        else:
            if nb_problems == 0:
                context["nb_problems_warnings"] = _(
                    '<strong class="green-bg">No problem</strong> found, and '
                )
            else:
                context["nb_problems_warnings"] = '<strong class="red-bg">'
                context["nb_problems_warnings"] += ngettext(
                    "{nb_problems} problem</strong> found, and ",
                    "{nb_problems} problems</strong> found, and ",
                    nb_problems,
                ).format(nb_problems=nb_problems)

            if nb_warnings == 0:
                context["nb_problems_warnings"] += _(
                    '<strong class="green-bg">no warning</strong> raised'
                )
            else:
                context["nb_problems_warnings"] += '<strong class="yellow-bg">'
                context["nb_problems_warnings"] += ngettext(
                    "{nb_warnings} warning</strong> raised",
                    "{nb_warnings} warnings</strong> raised",
                    nb_warnings,
                ).format(nb_warnings=nb_warnings)

        return context

    def render_to_response(self, context, **response_kwargs):
        if self.request.GET.get("fragment") == "checks":
            return render(
                self.request,
                "django_check_seo/fragment_checks.html",
                context,
            )
        return super().render_to_response(context, **response_kwargs)


def _base_urls(request, path_value=None, cms_page_id=None, language=None):
    """Build keywords_edit_url and keywords_cancel_url for the fragment."""
    base = request.path
    if cms_page_id is not None:
        cancel = (
            base
            + "?cms_page_id="
            + str(cms_page_id)
            + "&language="
            + quote(language or "", safe="")
            + "&show_form=0"
        )
    else:
        cancel = base + "?path=" + _quote_path(path_value or "") + "&show_form=0"
    return {
        "keywords_edit_url": base,
        "keywords_cancel_url": cancel,
    }


class KeywordsEditView(PermissionRequiredMixin, generic.View):
    """GET: return keywords form fragment (form or list). POST: save keywords and return list fragment + HX-Trigger."""

    permission_required = "django_check_seo.use_django_check_seo"

    def get(self, request):
        path_value = request.GET.get("path") or ""
        cms_page_id = request.GET.get("cms_page_id")
        if cms_page_id is not None:
            try:
                cms_page_id = int(cms_page_id)
            except (TypeError, ValueError):
                cms_page_id = None
        language = request.GET.get("language") or ""
        keywords = _get_keywords_for_page_identifier(
            path=path_value or None,
            cms_page_id=cms_page_id,
            language=language,
        )
        use_cms = cms_page_id is not None
        context = {
            "keywords": keywords,
            "current_page_path": path_value
            or (f"/cms-page-{cms_page_id}/" if use_cms else ""),
            "message_success": False,
            "keywords_text": ", ".join(keywords),
            "all_keywords": _get_all_keyword_names(),
            "use_cms_page_id": use_cms,
            "current_page_cms_id": cms_page_id if use_cms else None,
            "current_page_language": language if use_cms else None,
            **_base_urls(
                request,
                path_value=path_value,
                cms_page_id=cms_page_id,
                language=language,
            ),
        }
        return render(request, "django_check_seo/keywords_form.html", context)

    def post(self, request):
        path_value = (request.POST.get("path") or "").strip()
        cms_page_id = request.POST.get("cms_page_id")
        if cms_page_id is not None:
            try:
                cms_page_id = int(cms_page_id)
            except (TypeError, ValueError):
                cms_page_id = None
        language = (request.POST.get("language") or "").strip()
        page, path_display, use_cms = _get_or_create_page_for_edit(
            path=path_value if path_value else None,
            cms_page_id=cms_page_id,
            language=language,
        )
        action = (request.POST.get("action") or "").strip()

        if action == "update":
            old_name = (request.POST.get("old_keyword") or "").strip()
            new_name = (request.POST.get("new_keyword") or "").strip()
            if old_name and new_name:
                try:
                    old_kw = Keyword.objects.get(name=old_name)
                    new_kw, _ = Keyword.objects.get_or_create(
                        name=new_name, defaults={}
                    )
                    page.keywords.remove(old_kw)
                    page.keywords.add(new_kw)
                except Keyword.DoesNotExist:
                    pass
        elif action == "remove":
            name = (request.POST.get("keyword") or "").strip()
            if name:
                try:
                    kw = Keyword.objects.get(name=name)
                    page.keywords.remove(kw)
                except Keyword.DoesNotExist:
                    pass
        elif action == "add":
            name = (request.POST.get("keyword") or "").strip()
            if name:
                kw, _ = Keyword.objects.get_or_create(name=name, defaults={})
                page.keywords.add(kw)
        else:
            # bulk save (textarea)
            keywords_text = request.POST.get("keywords_text", "")
            names = _parse_keywords_text(keywords_text)
            keyword_objs = []
            for name in names:
                kw, _ = Keyword.objects.get_or_create(name=name.strip(), defaults={})
                keyword_objs.append(kw)
            page.keywords.set(keyword_objs)

        keywords = [kw.name for kw in page.keywords.all()]
        context = {
            "keywords": keywords,
            "current_page_path": path_display,
            "message_success": True,
            "keywords_text": "",
            "all_keywords": _get_all_keyword_names(),
            "use_cms_page_id": use_cms,
            "current_page_cms_id": page.cms_page_id if use_cms else None,
            "current_page_language": page.language if use_cms else None,
            **_base_urls(
                request,
                path_value=path_display,
                cms_page_id=page.cms_page_id if use_cms else None,
                language=page.language if use_cms else None,
            ),
        }
        response = render(request, "django_check_seo/keywords_form.html", context)
        response["HX-Trigger"] = "refreshChecks"
        return response



from bs4 import BeautifulSoup

from django_check_seo.checks import site
from django_check_seo.utils.keywords_discovery import meta_keywords

# Use ./launch_tests.sh to launch these tests.


html_content = """
<!doctype html>
<html>
    <head>
        <meta name="keywords" content="description,  title">
    </head>
</html>
"""


class init:
    def __init__(self):
        self.keywords = []
        self.problems = []
        self.warnings = []
        self.success = []

        self.soup = BeautifulSoup(html_content, features="lxml")
        self.full_url = "https://localhost/fake-url/title-of-the-page/"
        # populate class with data
        self.page_stats = site.Site(self.soup, self.full_url)


def test_keyword_url_importance():
    from django_check_seo.checks_list import check_keyword_url

    assert check_keyword_url.importance() == 1


def test_keyword_url_kw():
    from django_check_seo.checks_list import check_keyword_url, check_keywords

    site = init()

    meta_keywords(site)
    check_keywords.run(site)
    check_keyword_url.run(site)

    for success in site.success:
        if success.name == "Keywords found in URL":
            assert success.name == "Keywords found in URL"
            assert success.settings == "at least one"
            assert success.found == "title"
            assert success.searched_in == [
                'https://localhost/fake-url/<b class="good">title</b>-of-the-page/'
            ]
            assert (
                success.description
                == 'Keywords in URL will help your users understand the organisation of your website, and are a small ranking factor for Google. On the other hand, Bing guidelines advises to "<i>keep [your URL] clean and keyword rich when possible</i>".<br />Warning, Django Check SEO will try to find keywords in the URL without apostrophes ("pour-lenergie" will be found, but not "pour-l-energie").'
            )


def test_keyword_url_nokw():
    from django_check_seo.checks_list import check_keyword_url, check_keywords

    site = init()
    meta_keywords(site)
    check_keywords.run(site)
    site.full_url = "https://localhost/fake-url/notitle-of-the-page"
    check_keyword_url.run(site)

    for problem in site.problems:
        if problem.name == "No keyword in URL":
            assert problem.name == "No keyword in URL"
            assert problem.settings == "at least one"
            assert problem.found == "none"
            assert problem.searched_in == [
                "https://localhost/fake-url/notitle-of-the-page"
            ]


def test_keyword_url_nokw_root():
    from django_check_seo.checks_list import check_keyword_url, check_keywords

    site = init()
    meta_keywords(site)
    check_keywords.run(site)
    site.full_url = "https://localhost/"
    check_keyword_url.run(site)

    for success in site.success:
        if success.name == "Keywords found in URL":
            raise ValueError("We don't espect kw to be found in root URL")

    for problem in site.problems:
        if problem.name == "No keyword in URL":
            raise ValueError(
                "We shouldnt return a problem if a keyword is not found in root URL"
            )


def test_keyword_url_kws():
    from django_check_seo.checks_list import check_keyword_url, check_keywords

    site = init()
    site.soup.select('meta[name="keywords"]')[0]["content"] = "title,  page"

    meta_keywords(site)
    check_keywords.run(site)
    site.full_url = "https://localhost/fake-url/title-of-the-page"

    check_keyword_url.run(site)

    for success in site.success:
        if success.name == "Keywords found in URL":
            assert success.name == "Keywords found in URL"
            assert success.settings == "at least one"
            assert success.found == "title, page"
            assert success.searched_in == [
                'https://localhost/fake-url/<b class="good">title</b>-of-the-<b class="good">page</b>'
            ]


def test_keyword_url_kw_accented():
    from django_check_seo.checks_list import check_keyword_url, check_keywords

    site = init()
    site.soup.select('meta[name="keywords"]')[0]["content"] = "énergie"

    meta_keywords(site)
    check_keywords.run(site)
    site.full_url = "https://localhost/fake-url/title-of-energie"

    check_keyword_url.run(site)

    for success in site.success:
        if success.name == "Keywords found in URL":
            assert success.name == "Keywords found in URL"
            assert success.settings == "at least one"
            assert success.found == "énergie"
            assert success.searched_in == [
                'https://localhost/fake-url/title-of-<b class="good">energie</b>'
            ]


def test_keyword_url_kws_accented():
    from django_check_seo.checks_list import check_keyword_url, check_keywords

    site = init()
    site.soup.select('meta[name="keywords"]')[0]["content"] = "énergie,  éééé"

    meta_keywords(site)
    check_keywords.run(site)
    site.full_url = "https://localhost/fake-url/title-eeee-energie"

    check_keyword_url.run(site)

    for success in site.success:
        if success.name == "Keywords found in URL":
            assert success.name == "Keywords found in URL"
            assert success.settings == "at least one"
            assert success.found == "énergie, éééé"
            assert success.searched_in == [
                'https://localhost/fake-url/title-<b class="good">eeee</b>-<b class="good">energie</b>'
            ]


def test_keyword_url_kws_accented_unaccented():
    from django_check_seo.checks_list import check_keyword_url, check_keywords

    site = init()
    site.soup.select('meta[name="keywords"]')[0]["content"] = "énergie,  title"

    meta_keywords(site)
    check_keywords.run(site)
    site.full_url = "https://localhost/fake-url/title-of-energie"

    check_keyword_url.run(site)

    for success in site.success:
        if success.name == "Keywords found in URL":
            assert success.name == "Keywords found in URL"
            assert success.settings == "at least one"
            assert success.found == "énergie, title"
            assert success.searched_in == [
                'https://localhost/fake-url/<b class="good">title</b>-of-<b class="good">energie</b>'
            ]



# Use ./launch_tests.sh to launch these tests.

from bs4 import BeautifulSoup

from django_check_seo.checks import site
from django_check_seo.utils.keywords_discovery import meta_keywords

html_content = """
<!doctype html>
<html>
    <head>
        <meta name="keywords" content="description,  title">
        <title>Title of the page</title>
    </head>
</html>
"""


class settings:
    def __init__(self):
        self.DJANGO_CHECK_SEO_SETTINGS = {
            "meta_title_length": [30, 60],
        }


class init:
    def __init__(self):
        self.keywords = []
        self.problems = []
        self.warnings = []
        self.success = []
        self.settings = settings()
        self.soup = BeautifulSoup(html_content, features="lxml")
        self.full_url = "https://localhost/fake-url/title-of-the-page/"
        # populate class with data
        self.page_stats = site.Site(self.soup, self.full_url)


def test_title_importance():
    from django_check_seo.checks_list import check_title

    assert check_title.importance() == 1


def test_title_notitle():
    from django_check_seo.checks_list import check_title

    site = init()
    site.soup.find("title").decompose()
    check_title.run(site)

    for problem in site.problems:
        if problem.name == "No meta title tag":
            assert problem.name == "No meta title tag"
            assert problem.settings == "one"
            assert problem.found == "none"
            assert problem.searched_in == []
            assert (
                problem.description
                == "Titles tags are ones of the most important things to add to your pages, sinces they are the main text displayed on result search pages."
            )


def test_title_short():
    from django_check_seo.checks_list import check_title

    site = init()
    check_title.run(site)

    for problem in site.problems:
        if problem.name == "Meta title tag is too short":
            assert problem.name == "Meta title tag is too short"
            assert problem.settings == "more than 30"
            assert problem.found == 17
            assert problem.searched_in == ["Title of the page"]
            assert (
                problem.description
                == "Meta titles tags need to describe the content of the page, and need to contain at least a few words."
            )


def test_title_long():
    from django_check_seo.checks_list import check_title

    site = init()
    site.soup.find(
        "title"
    ).string = "Lorem ipsum dolor sit amet, consectetur adipiscing elit. Etiam pretium in quam sit amet egestas. Nullam cursus mi eu mauris congue laoreet."
    check_title.run(site)

    for warning in site.warnings:
        if warning.name == "Meta title tag is too long":
            assert warning.name == "Meta title tag is too long"
            assert warning.settings == "less than 60"
            assert warning.found == 139
            assert warning.searched_in == [
                'Lorem ipsum dolor sit amet, consectetur adipiscing elit. Eti<b class="problem">am pretium in quam sit amet egestas. Nullam cursus mi eu mauris congue laoreet.</b>'
            ]
            assert (
                warning.description
                == "Only the first ~55-60 chars are displayed on modern search engines results. Writing a longer meta title is not really required and can lead to make the user miss informations."
            )


def test_title_okay():
    from django_check_seo.checks_list import check_title

    site = init()
    site.soup.find("title").string = "This is a title with a good length."
    check_title.run(site)

    for success in site.success:
        if success.name == "Meta title tag have a good length":
            assert success.name == "Meta title tag have a good length"
            assert success.settings == "more than 30"
            assert success.found == 35
            assert success.searched_in == [
                'This is a title with a good le<b class="good">ngth.</b>'
            ]
            assert (
                success.description
                == "Meta titles tags need to describe the content of the page."
            )


def test_titles():
    import copy

    from django_check_seo.checks_list import check_title

    site = init()
    site.soup.head.append(copy.copy(site.soup.find("title")))
    check_title.run(site)

    for problem in site.problems:
        if problem.name == "Too much meta title tags":
            assert problem.name == "Too much meta title tags"
            assert problem.settings == "only one"
            assert problem.found == 2
            assert problem.searched_in == ["Title of the page", "Title of the page"]
            assert (
                problem.description
                == "Only the first meta title tag will be displayed on the tab space on your browser, and only one meta title tag will be displayed on the search results pages."
            )


def test_title_nokw():
    from django_check_seo.checks_list import check_keywords, check_title

    site = init()
    meta_keywords(site)
    check_keywords.run(site)

    site.soup.find("title").string = "There is notitle on this page."
    check_title.run(site)

    for problem in site.problems:
        if problem.name == "Meta title tag do not contain any keyword":
            assert problem.name == "Meta title tag do not contain any keyword"
            assert problem.settings == "at least one"
            assert problem.found == "none"
            assert problem.searched_in == ["there is notitle on this page."]
            assert (
                problem.description
                == "Meta titles tags need to contain at least one keyword, since they are one of the most important content of the page for search engines."
            )


def test_title_kw():
    from django_check_seo.checks_list import check_keywords, check_title

    site = init()
    meta_keywords(site)
    check_keywords.run(site)

    check_title.run(site)

    for success in site.success:
        if success.name == "Keywords found in meta title tag":
            assert success.name == "Keywords found in meta title tag"
            assert success.settings == "at least one"
            assert success.found == "title"
            assert success.searched_in == ['<b class="good">title</b> of the page']
            assert (
                success.description
                == "Meta titles tags need to contain at least one keyword, since they are one of the most important content of the page for search engines."
            )


def test_title_kws():
    from django_check_seo.checks_list import check_keywords, check_title

    site = init()
    meta_keywords(site)
    check_keywords.run(site)

    site.soup.find("title").string = "This title have a description."
    check_title.run(site)

    for success in site.success:
        if success.name == "Keywords found in meta title tag":
            assert success.name == "Keywords found in meta title tag"
            assert success.settings == "at least one"
            assert success.found == "description, title"
            assert success.searched_in == [
                'this <b class="good">title</b> have a <b class="good">description</b>.'
            ]



# -*- coding: utf-8 -*-
"""
Keywords discovery functions.

Each function has the signature (site) -> None and must set site.keywords
to a list of strings. Used when DJANGO_CHECK_SEO_KEYWORDS_DISCOVERY_METHOD
points to one of these functions.
"""

from urllib.parse import urlparse

from django_check_seo.models import Page


def _normalize_path(path):
    """Ensure path starts with / for lookup."""
    path = path or "/"
    if not path.startswith("/"):
        path = "/" + path
    return path


def get_cms_page_id_language_for_path(path):
    """
    Resolve path to (cms_page_id, language) via Django CMS PageUrl if available.

    Returns (page_id, language) or (None, None) when not a CMS page (e.g. apphook)
    or when django CMS is not installed.
    """
    path = _normalize_path(path)
    try:
        from cms.models import PageUrl

        page_url = (
            PageUrl.objects.filter(path=path).values_list("page_id", "language").first()
        )
        if page_url:
            return (page_url[0], page_url[1] or "")
        return (None, None)
    except Exception:
        # ImportError when cms is not installed; RuntimeError etc. when cms is not in INSTALLED_APPS
        return (None, None)


def meta_keywords(site):
    """Populate site.keywords from the meta name="keywords" content attribute."""
    site.keywords = []
    meta = site.soup.find_all("meta")
    for tag in meta:
        if (
            "name" in tag.attrs
            and tag.attrs["name"] == "keywords"
            and "content" in tag.attrs
            and tag.attrs["content"].strip() != ""
        ):
            site.keywords = list(
                map(str.strip, tag.attrs["content"].strip().split(","))
            )
            break


def model_keywords(site):
    """Populate site.keywords from Page/Keyword models for the current path."""
    site.keywords = []
    path = urlparse(site.full_url).path or "/"
    path = _normalize_path(path)
    try:
        page = Page.objects.get(path=path)
        site.keywords = [kw.name for kw in page.keywords.all()]
    except Page.DoesNotExist:
        pass


def django_cms_model_keywords(site):
    """
    Populate site.keywords from Page/Keyword models, using (cms_page_id, language)
    when the path resolves to a Django CMS page, else fallback to path (apphooks).
    """
    site.keywords = []
    path = urlparse(site.full_url).path or "/"
    path = _normalize_path(path)

    cms_page_id, language = get_cms_page_id_language_for_path(path)
    if cms_page_id is not None:
        page = Page.objects.filter(cms_page_id=cms_page_id, language=language).first()
        if page:
            site.keywords = [kw.name for kw in page.keywords.all()]
            return

    # Fallback: path-only (e.g. apphook or non-CMS)
    try:
        page = Page.objects.get(path=path)
        site.keywords = [kw.name for kw in page.keywords.all()]
    except Page.DoesNotExist:
        pass



# Standard Library
from urllib.parse import urlparse

import unidecode
from django.conf import settings
from django.conf.global_settings import LANGUAGES
from django.utils.translation import gettext as _
from django.utils.translation import pgettext_lazy

# Local application / specific library imports
from ..checks import custom_list, utils


def importance():
    """Scripts with higher importance will be executed first.

    Returns:
        int -- Importance of the script.
    """
    return 1


def run(site):
    """Check presence of keywords in url.

    Arguments:
        site {Site} -- Structure containing a good amount of resources from the targeted webpage.
    """

    no_keyword = custom_list.CustomList(
        name=_("No keyword in URL"),
        settings=pgettext_lazy("masculin", "at least one"),
        found=pgettext_lazy("masculin", "none"),
        description=_(
            'Keywords in URL will help your users understand the organisation of your website, and are a small ranking factor for Google. On the other hand, Bing guidelines advises to "<i>keep [your URL] clean and keyword rich when possible</i>".<br />Warning, Django Check SEO will try to find keywords in the URL without apostrophes ("pour-lenergie" will be found, but not "pour-l-energie").'
        ),
    )

    enough_keyword = custom_list.CustomList(
        name=_("Keywords found in URL"),
        settings=pgettext_lazy("masculin", "at least one"),
        found="",
        description=no_keyword.description,
    )

    # root url may contain str like "/fr/" or "/en/" if i18n is activated
    url_path = urlparse(site.full_url, "/").path

    # list of languages from django LANGUAGES list: ['fr', 'en', 'br', 'ia', ...]
    languages_list = [i[0] for i in LANGUAGES]

    # do not check keywords in url for root URL
    if (
        (settings.USE_I18N and url_path.replace("/", "") in languages_list)
        or url_path == "/"
        or not url_path
    ):
        return

    full_url = site.full_url.lower()
    occurrence = []
    accented_occurrences = 0

    for keyword in site.keywords:
        keyword = keyword.lower().replace(" ", "-")

        # remove apostrophes as they are removed from URLs
        keyword = keyword.replace("'", "").replace("’", "")

        keyword_unnaccented = unidecode.unidecode(keyword)

        nb_occurrences = utils.count_keyword_occurrences(
            keyword, full_url, for_url=True
        )
        if nb_occurrences == 0:
            # retry with unnaccented kw
            accented_occurrences = utils.count_keyword_occurrences(
                keyword_unnaccented, full_url, for_url=True
            )
        occurrence.append(nb_occurrences + accented_occurrences)

        if nb_occurrences > 0:
            full_url = full_url.replace(keyword, '<b class="good">' + keyword + "</b>")
            if enough_keyword.found != "":
                enough_keyword.found += ", "
            enough_keyword.found += keyword
            nb_occurrences = 0

        if accented_occurrences > 0:
            full_url = full_url.replace(
                keyword_unnaccented, '<b class="good">' + keyword_unnaccented + "</b>"
            )
            if enough_keyword.found != "":
                enough_keyword.found += ", "
            enough_keyword.found += keyword
            accented_occurrences = 0

    if not any(i > 0 for i in occurrence):
        no_keyword.searched_in = [full_url]
        site.problems.append(no_keyword)
    else:
        enough_keyword.searched_in = [full_url]
        site.success.append(enough_keyword)



# Generated by Django 4.2.9 on 2024-01-25 12:38

from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="DjangoCheckSEOPermissions",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
            options={
                "permissions": (
                    (
                        "use_django_check_seo",
                        "View the Check SEO button (if using Django CMS) and use Django Check SEO",
                    ),
                ),
                "managed": False,
                "default_permissions": (),
            },
        ),
    ]



from django.db import models
from django.db.models import Q

try:
    from django.utils.translation import ugettext_lazy as _
except ImportError:
    from django.utils.translation import gettext_lazy as _


class Keyword(models.Model):
    """Keyword used for SEO checks. Unique by name."""

    name = models.CharField(_("name"), max_length=255, unique=True)

    class Meta:
        verbose_name = _("keyword")
        verbose_name_plural = _("keywords")

    def __str__(self):
        return self.name

    def pages_using(self):
        return ", ".join(page.get_display_identifier() for page in self.pages.all())

    pages_using.short_description = _("Pages using this keyword")


class Page(models.Model):
    """
    Page identified by path (e.g. /fr/ma-page/) and/or by (cms_page_id, language).
    M2M to Keyword. Hybrid: use path for apphooks/non-CMS URLs,
    (cms_page_id, language) for CMS pages.
    """

    path = models.CharField(
        _("path"), max_length=500, null=True, blank=True, db_index=True
    )
    cms_page_id = models.PositiveIntegerField(
        _("CMS page id"), null=True, blank=True, db_index=True
    )
    language = models.CharField(_("language"), max_length=10, null=True, blank=True)
    keywords = models.ManyToManyField(
        Keyword,
        related_name="pages",
        blank=True,
        verbose_name=_("keywords"),
    )

    class Meta:
        verbose_name = _("page")
        verbose_name_plural = _("pages")
        constraints = [
            models.UniqueConstraint(
                fields=["path"],
                condition=Q(path__isnull=False) & ~Q(path=""),
                name="dcs_unique_page_path_when_set",
            ),
            models.UniqueConstraint(
                fields=["cms_page_id", "language"],
                condition=Q(cms_page_id__isnull=False),
                name="dcs_unique_cms_page_language",
            ),
        ]

    def __str__(self):
        return self.get_display_identifier()

    def get_display_identifier(self):
        """Return path if set, else 'cms_page_id=X (lang)' for display."""
        if self.path:
            return self.path
        if self.cms_page_id is not None:
            lang = self.language or ""
            return f"cms_page_id={self.cms_page_id} ({lang})"
        return ""

    def nb_keywords(self):
        return self.keywords.count()

    nb_keywords.short_description = _("Number of keywords")


class DjangoCheckSEOPermissions(models.Model):
    class Meta:
        managed = False
        default_permissions = ()
        permissions = (
            (
                "use_django_check_seo",
                _(
                    "View the Check SEO button (if using Django CMS) and use Django Check SEO"
                ),
            ),
        )



# Third party
from django.utils.translation import gettext as _
from django.utils.translation import pgettext_lazy

# Local application / specific library imports
from ..checks import custom_list, utils


def importance():
    """Scripts with higher importance will be executed first.

    Returns:
        int -- Importance of the script.
    """
    return 1


def run(site):
    """Checks that h2 tags exist, and that at least one h2 contains at least one keyword.

    Arguments:
        site {Site} -- Structure containing a good amount of resources from the targeted webpage.
    """

    no_h2 = custom_list.CustomList(
        name=_("No h2 tag"),
        settings=_("at least one"),
        found=pgettext_lazy("masculin", "none"),
        description=_(
            "H2 tags are useful because they are explored by search engines and can help them understand the subject of your page."
        ),
    )

    enough_h2 = custom_list.CustomList(
        name=_("H2 tags were found"),
        settings=pgettext_lazy("feminin", "at least one"),
        description=no_h2.description,
    )

    no_keywords = custom_list.CustomList(
        name=_("No keyword in h2 tags"),
        settings=pgettext_lazy("masculin", "at least one"),
        found=pgettext_lazy("masculin", "none"),
        description=_(
            "Google uses h2 tags to better understand the subjects of your page."
        ),
    )

    enough_keywords = custom_list.CustomList(
        name=_("Keyword found in h2 tags"),
        settings=pgettext_lazy("masculin", "at least one"),
        found="",
        description=no_keywords.description,
    )

    h2 = site.soup.find_all("h2")
    if not h2:
        site.warnings.append(no_h2)
    else:
        enough_h2.found = len(h2)
        enough_h2.searched_in = [utils.get_heading_text(t) for t in h2]
        site.success.append(enough_h2)

        occurrence = []
        h2_kw = []

        # for each h2...
        for single_h2_tag in h2:
            text = utils.get_heading_text(single_h2_tag).lower()
            highlighted, occ, found = utils.highlight_keywords_in_text(
                text, site.keywords, normalize_apostrophes_flag=True
            )
            occurrence.extend(occ)
            h2_kw.append(highlighted)
            if found:
                enough_keywords.found += (", " if enough_keywords.found else "") + found

        # if no keyword is found in h2
        if not any(i > 0 for i in occurrence):
            no_keywords.searched_in = [t.text for t in h2]
            no_keywords.found = pgettext_lazy("masculin", "none")
            site.warnings.append(no_keywords)
        else:
            enough_keywords.searched_in = h2_kw
            site.success.append(enough_keywords)



# Generated by Django 4.2 on 2025-03-10

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_check_seo", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="Keyword",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(max_length=255, unique=True, verbose_name="name"),
                ),
            ],
            options={
                "verbose_name": "keyword",
                "verbose_name_plural": "keywords",
            },
        ),
        migrations.CreateModel(
            name="Page",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "path",
                    models.CharField(max_length=500, unique=True, verbose_name="path"),
                ),
                (
                    "keywords",
                    models.ManyToManyField(
                        blank=True,
                        related_name="pages",
                        to="django_check_seo.keyword",
                        verbose_name="keywords",
                    ),
                ),
            ],
            options={
                "verbose_name": "page",
                "verbose_name_plural": "pages",
            },
        ),
    ]



# Use ./launch_tests.sh to launch these tests.

from bs4 import BeautifulSoup

from django_check_seo.checks import site

html_content = """
<!doctype html>
<html>
    <head>
        <meta name="keywords" content="description,  title">
    </head>
    <a href="/link1">Heyy !</a>
</html>
"""


class settings:
    def __init__(self):
        self.DJANGO_CHECK_SEO_SETTINGS = {
            "internal_links": 1,
            "external_links": 1,
            "max_link_depth": 4,
            "max_url_length": 70,
        }


class init:
    def __init__(self):
        self.keywords = []
        self.problems = []
        self.warnings = []
        self.success = []
        self.settings = settings()
        self.soup = BeautifulSoup(html_content, features="lxml")
        self.content = self.soup.find_all("body")
        self.full_url = "https://localhost/fake-url/title-of-the-page/"
        # populate class with data
        self.page_stats = site.Site(self.soup, self.full_url)


class patch_get_content_class:
    def __init__(self):
        self.domain = "https://localhost/"


def test_link_importance():
    from django_check_seo.checks_list import check_links

    assert check_links.importance() == 1


def test_links_internal_noexternal(monkeypatch):
    from django.contrib.sites.models import Site

    from django_check_seo.checks_list import check_links

    site = init()
    monkeypatch.setattr(Site.objects, "get_current", patch_get_content_class)
    check_links.run(site)

    for warning in site.warnings:
        if warning.name == "Not enough external links":
            assert warning.name == "Not enough external links"
            assert warning.settings == "at least 1"
            assert warning.found == 0
            assert warning.searched_in == []
            assert (
                warning.description
                == "External links help your users to check your topic and can save them from having to do additional research."
            )

    for success in site.success:
        if success.name == "Internal links were found":
            assert success.name == "Internal links were found"
            assert success.settings == "at least 1"
            assert success.found == 1
            assert success.searched_in == ['<a href="/link1">Heyy !</a>']
            assert (
                success.description
                == "Internal links are useful because they can give the structure of your website to search engines, so they can create a hierarchy of your pages."
            )


def test_links_nointernal_external(monkeypatch):
    from django.contrib.sites.models import Site

    from django_check_seo.checks_list import check_links

    site = init()
    site.soup.find("a")["href"] = "https://github.com/kapt-labs/django-check-seo"
    monkeypatch.setattr(Site.objects, "get_current", patch_get_content_class)
    check_links.run(site)

    for warning in site.warnings:
        if warning.name == "Not enough internal links":
            assert warning.name == "Not enough internal links"
            assert warning.settings == "at least 1"
            assert warning.found == 0
            assert warning.searched_in == []
            assert (
                warning.description
                == "Internal links are useful because they can give the structure of your website to search engines, so they can create a hierarchy of your pages."
            )

    for success in site.success:
        if success.name == "External links were found":
            assert success.name == "External links were found"
            assert success.settings == "at least 1"
            assert success.found == 1
            assert success.searched_in == [
                '<a href="https://github.com/kapt-labs/django-check-seo">Heyy !</a>'
            ]
            assert (
                success.description
                == "External links help your users to check your topic and can save them from having to do additional research."
            )


def test_links_internal_img(monkeypatch):
    from django.contrib.sites.models import Site

    from django_check_seo.checks_list import check_links

    site = init()
    site.soup.find("a").decompose()
    site.soup.body.append(
        (
            BeautifulSoup(
                '<a href="/"><img src="none" alt="Hey!!!" /></a>',
                features="lxml",
            )
        )
    )
    monkeypatch.setattr(Site.objects, "get_current", patch_get_content_class)
    check_links.run(site)

    for success in site.success:
        if success.name == "Internal links were found":
            assert success.name == "Internal links were found"
            assert success.settings == "at least 1"
            assert success.found == 1
            assert success.searched_in == ['<a href="/">Hey!!! (&lt;img&gt;)</a>']


def test_links_internal_img_no_alt_tag(monkeypatch):
    from django.contrib.sites.models import Site

    from django_check_seo.checks_list import check_links

    site = init()
    site.soup.find("a").decompose()
    site.soup.body.append(
        (
            BeautifulSoup(
                '<a href="/"><img src="none" /></a>',
                features="lxml",
            )
        )
    )
    monkeypatch.setattr(Site.objects, "get_current", patch_get_content_class)
    check_links.run(site)

    for success in site.success:
        if success.name == "Internal links were found":
            assert success.name == "Internal links were found"
            assert success.settings == "at least 1"
            assert success.found == 1
            assert success.searched_in == ['<a href="/">&lt;img src="none"/&gt;</a>']


def test_links_internal_empty(monkeypatch):
    from django.contrib.sites.models import Site

    from django_check_seo.checks_list import check_links

    site = init()
    site.soup.find("a").decompose()
    site.soup.body.append(
        (
            BeautifulSoup(
                '<a href="/"></a>',
                features="lxml",
            )
        )
    )
    monkeypatch.setattr(Site.objects, "get_current", patch_get_content_class)
    check_links.run(site)

    for success in site.success:
        if success.name == "Internal links were found":
            assert success.name == "Internal links were found"
            assert success.settings == "at least 1"
            assert success.found == 1
            assert success.searched_in == ['<a href="/">[no content]</a>']



"""
Shared utilities for SEO checks.
"""

import re

# Word boundaries for keyword matching: allow symbols (e.g. ®) before/after keyword,
# not letters (e.g. avoid "art" in "artisanat").
# See this issue for more details:
# https://github.com/kapt-labs/django-check-seo/issues/65
# (previously https://github.com/kapt-labs/django-check-seo/issues/38)
_KEYWORD_BOUNDARY_LEFT = r"(^| |\n|,|\.|!|\?|[^\w'])"
_KEYWORD_BOUNDARY_RIGHT = r"($| |\n|,|\.|!|\?|[^\w'])"
_KEYWORD_BOUNDARY_LEFT_URL = r"(^| |\n|,|\.|!|\?|/|-|[^\w'])"
_KEYWORD_BOUNDARY_RIGHT_URL = r"($| |\n|,|\.|!|\?|/|-|[^\w'])"


def normalize_apostrophes(s):
    """Replace straight apostrophe by typographic one for consistent keyword matching."""
    return s.replace("'", "\u2019")  # U+2019 right single quotation mark


def get_heading_text(tag):
    """Return heading text, or img alt if heading only contains an image with alt."""
    if not tag.text and tag.find("img", {"alt": True}):
        return tag.find("img")["alt"]
    return tag.text


def highlight_keywords_in_text(
    text, keywords, normalize_apostrophes_flag=False, for_url=False
):
    """Count keyword occurrences, highlight them with <b class="good">, return found string.

    Arguments:
        text: The text to search in (assumed already lowercased by caller).
        keywords: Iterable of keyword strings.
        normalize_apostrophes_flag: If True, normalize apostrophes in text and keywords.
        for_url: If True, use URL boundaries for counting.

    Returns:
        tuple: (highlighted_text, occurrence_list, found_string)
        - highlighted_text: text with keywords wrapped in <b class="good">...</b>
        - occurrence_list: one int per keyword (count for that keyword)
        - found_string: comma-separated list of keywords that were found
    """
    if normalize_apostrophes_flag:
        text = normalize_apostrophes(text)
    occurrence_list = []
    found_kw = []
    for keyword in keywords:
        kw = keyword.lower()
        if normalize_apostrophes_flag:
            kw = normalize_apostrophes(kw)
        count = count_keyword_occurrences(kw, text, for_url=for_url)
        occurrence_list.append(count)
        if count > 0:
            text = text.replace(kw, '<b class="good">' + kw + "</b>")
            found_kw.append(kw)
    return (text, occurrence_list, ", ".join(found_kw))


def count_keyword_occurrences(keyword, text, for_url=False):
    """Count occurrences of a keyword in text with word boundaries (optional plural s).

    Boundaries include [^\\w'] so symbols like ® before/after the keyword are accepted,
    while letters (e.g. "artisanat" for keyword "art") are not.

    Arguments:
        keyword: The keyword to search (will be used literally in regex).
        text: The text to search in.
        for_url: If True, use URL boundaries (adds / and - as separators).

    Returns:
        int: Number of matches.
    """
    if for_url:
        left, right = _KEYWORD_BOUNDARY_LEFT_URL, _KEYWORD_BOUNDARY_RIGHT_URL
    else:
        left, right = _KEYWORD_BOUNDARY_LEFT, _KEYWORD_BOUNDARY_RIGHT
    # keyword is used literally (caller normalizes apostrophes etc.)
    pattern = left + keyword + r"s?" + right
    return len(re.findall(pattern, text))



# Third party
from django.utils.translation import gettext as _
from django.utils.translation import pgettext_lazy

# Local application / specific library imports
from ..checks import custom_list


def importance():
    """Scripts with higher importance will be executed first.

    Returns:
        int -- Importance of the script.
    """
    return 5


def run(site):
    """Validate that at least one keyword is defined for this page.

    site.keywords is expected to be already populated by the configured
    keywords discovery method (meta or model). This check only validates
    presence and reports success or problem with generic labels.

    Arguments:
        site {Site} -- Structure containing a good amount of resources from the targeted webpage.
    """

    no_keywords = custom_list.CustomList(
        name=_("No keywords defined for this page"),
        settings=pgettext_lazy("masculin", "at least one"),
        found=pgettext_lazy("masculin", "none"),
        description=_(
            "Django-check-seo uses the keywords to check all other tests related to them. "
            "A series of problems and warnings are related to keywords, and will therefore "
            "systematically be activated if the keywords are not filled in."
        ),
    )

    keywords_found = custom_list.CustomList(
        name=_("Keywords found"),
        settings=pgettext_lazy("masculin", "at least one"),
        found="",
        description=no_keywords.description,
    )

    if not site.keywords or len(site.keywords) == 0:
        site.problems.append(no_keywords)
        return

    keywords_found.found = len(site.keywords)
    keywords_found.searched_in = site.keywords
    site.success.append(keywords_found)



# Use ./launch_tests.sh to launch these tests.

from bs4 import BeautifulSoup

from django_check_seo.checks import site
from django_check_seo.utils.keywords_discovery import meta_keywords

html_content = """
<!doctype html>
<html>
    <head>
        <meta name="keywords" content="description,  title">
    </head>
</html>
"""


class init:
    def __init__(self):
        self.keywords = []
        self.problems = []
        self.warnings = []
        self.success = []

        self.soup = BeautifulSoup(html_content, features="lxml")
        self.full_url = "https://localhost/fake-url/title-of-the-page/"
        # populate class with data
        self.page_stats = site.Site(self.soup, self.full_url)


def test_keyword_importance():
    from django_check_seo.checks_list import check_keywords

    assert check_keywords.importance() == 5


def test_keyword_kw():
    from django_check_seo.checks_list import check_keywords

    site_obj = init()
    meta_keywords(site_obj)
    check_keywords.run(site_obj)

    for success in site_obj.success:
        if success.name == "Keywords found":
            assert success.name == "Keywords found"
            assert success.settings == "at least one"
            assert success.found == 2
            assert success.searched_in == ["description", "title"]
            assert "Django-check-seo uses the keywords" in success.description
            return
    assert False, "Expected 'Keywords found' in success"


def test_keyword_nokw():
    from django_check_seo.checks_list import check_keywords

    site_obj = init()
    site_obj.soup.select('meta[name="keywords"]')[0]["content"] = ""
    meta_keywords(site_obj)
    check_keywords.run(site_obj)

    for problem in site_obj.problems:
        if problem.name == "No keywords defined for this page":
            assert problem.name == "No keywords defined for this page"
            assert problem.settings == "at least one"
            assert problem.found == "none"
            assert problem.searched_in == []
            assert "Django-check-seo uses the keywords" in problem.description
            return
    assert False, "Expected 'No keywords defined for this page' in problems"


def test_meta_keywords_discovery():
    """meta_keywords(site) populates site.keywords from meta tag."""
    site_obj = init()
    assert site_obj.keywords == []
    meta_keywords(site_obj)
    assert site_obj.keywords == ["description", "title"]


def test_meta_keywords_discovery_empty():
    """meta_keywords leaves site.keywords empty when meta is missing or empty."""
    site_obj = init()
    site_obj.soup.select('meta[name="keywords"]')[0]["content"] = ""
    meta_keywords(site_obj)
    assert site_obj.keywords == []



# Third party
from django.contrib.admin.views.decorators import staff_member_required
from django.urls import path

# Local application / specific library imports
from . import views

urlpatterns = [
    path("", staff_member_required(views.IndexView.as_view()), name="Index"),
    path(
        "keywords-edit/",
        staff_member_required(views.KeywordsEditView.as_view()),
        name="keywords_edit",
    ),
]



# Third party
from django.utils.translation import gettext as _
from django.utils.translation import pgettext_lazy

# Local application / specific library imports
from ..checks import custom_list, utils


def importance():
    """Scripts with higher importance will be executed first.

    Returns:
        int -- Importance of the script.
    """
    return 1


def run(site):
    """Checks that at least one keyword is included in the first paragraph.

    Arguments:
        site {Site} -- Structure containing a good amount of resources from the targeted webpage.
    """

    no_keywords = custom_list.CustomList(
        name=_("No keyword in first paragraph"),
        settings=_("before {settings} words").format(
            settings=site.settings.DJANGO_CHECK_SEO_SETTINGS["keywords_in_first_words"]
        ),
        found="",
        description=_(
            "The reader will be relieved to find one of his keywords in the first paragraph of your page, and the same logic applies to Google, which will consider the content more relevant."
        ),
    )

    first_words = site.content_text.lower().split()[
        : site.settings.DJANGO_CHECK_SEO_SETTINGS["keywords_in_first_words"]
    ]
    # check text and not list of words in order to find keywords that looks like "this is a keyword" in text "words this is a keyword words"
    first_words = " ".join(first_words)

    highlighted, occurrence, found = utils.highlight_keywords_in_text(
        first_words, site.keywords, normalize_apostrophes_flag=True
    )

    no_keywords.searched_in = [highlighted]

    # no keyword was found in first paragraph
    if not any(i > 0 for i in occurrence):
        no_keywords.found = pgettext_lazy("masculin", "none")
        site.problems.append(no_keywords)

    else:
        no_keywords.name = _("Keywords found in first paragraph")
        no_keywords.found = found
        site.success.append(no_keywords)



# Third party
import bs4
from django.utils.translation import gettext as _

# Local application / specific library imports
from ..checks import custom_list


def importance():
    """Scripts with higher importance will be executed first.

    Returns:
        int -- Importance of the script.
    """
    return 1


def run(site):
    """Checks that each image has a filled alt tag.

    Arguments:
        site {Site} -- Structure containing a good amount of resources from the targeted webpage.
    """

    lack_alt = custom_list.CustomList(
        name=_("Images lack alt tag"),
        settings=_("all images"),
        found="",
        description=_(
            'Your images should have an alt tag, because it improves accessibility for visually impaired people.<br />But "<i>sometimes there is non-text content that really is not meant to be seen or understood by the user</i>" (<a href="https://www.w3.org/WAI/WCAG21/Understanding/non-text-content.html">WCAG</a>). For this kind of non-text content, you can leave your alt tag empty.<br />The name of the file is important too, because it helps Google understand what your image is about. For example, you could rename a file named "IMG0001.jpg" to "tree_with_a_bird.jpg".'
        ),
    )

    enough_alt = custom_list.CustomList(
        name=_("Images have alt tag"),
        settings=_("all images"),
        found="",
        description=lack_alt.description,
    )

    images = bs4.element.ResultSet(None)

    for c in site.content:
        images += c.find_all("img")

    warning = 0
    imgs = []

    for image in images:
        img_str = image.attrs["src"].split("/")[-1]
        if (
            "alt" not in image.attrs
            or image.attrs["alt"] == "None"
            or image.attrs["alt"] == ""
        ):
            warning += 1

            # bold without alt tag content
            if image.attrs["src"] != "":
                imgs.append(
                    '<b><u class="problem"><a target="_blank" href="'
                    + image.attrs["src"]
                    + '">'
                    + img_str
                    + "</a></u></b>"
                )
            # bold without alt tag content & without src content (dead img ?)
            else:
                imgs.append("<b><u>unknown image</u></b>")

        # normal with alt tag content
        else:
            imgs.append(
                '<a target="_blank" href="'
                + image.attrs["src"]
                + '">'
                + img_str
                + "</a> ("
                + image.attrs["alt"]
                + ")"
            )

    if warning > 0:
        lack_alt.found = warning
        lack_alt.searched_in = imgs
        site.warnings.append(lack_alt)
    else:
        if len(images) > 0:
            enough_alt.found = len(images)
            enough_alt.searched_in = imgs
            site.success.append(enough_alt)



# coding: utf-8

# Use ./launch_tests.sh to launch these tests.

from bs4 import BeautifulSoup

from django_check_seo.checks import site

html_content = """
<!doctype html>
<html>
    <body>
        <img src="none" alt="heyy" />
    </body>
</html>
"""


class init:
    def __init__(self):
        self.keywords = []
        self.problems = []
        self.warnings = []
        self.success = []

        self.soup = BeautifulSoup(html_content, features="lxml")
        self.content = self.soup.find_all("body")
        self.full_url = "https://localhost/fake-url/title-of-the-page/"
        # populate class with data
        self.page_stats = site.Site(self.soup, self.full_url)


def test_images_importance():
    from django_check_seo.checks_list import check_images

    assert check_images.importance() == 1


def test_image_okay():
    from django_check_seo.checks_list import check_images

    site = init()

    check_images.run(site)

    for success in site.success:
        if success.name == "Img have alt tag":
            assert success.name == "Img have alt tag"
            assert success.settings == "all images"
            assert success.found == 1
            assert success.searched_in == [
                '<a target="_blank" href="none">image</a> (heyy)'
            ]
            assert (
                success.description
                == 'Your images should always have an alt tag, because it improves accessibility for visually impaired people. The name of the file is important too, because it helps Google understand what your image is about. For example, you could rename a file named "IMG0001.jpg" to "tree_with_a_bird.jpg".'
            )


def test_images_okay():
    import copy

    from django_check_seo.checks_list import check_images

    site = init()

    site.soup.body.append(copy.copy(site.soup.find("img")))
    check_images.run(site)

    for success in site.success:
        if success.name == "Img have alt tag":
            assert success.name == "Img have alt tag"
            assert success.settings == "all images"
            assert success.found == 2
            assert success.searched_in == [
                '<a target="_blank" href="none">image</a> (heyy)',
                '<a target="_blank" href="none">image</a> (heyy)',
            ]
            assert (
                success.description
                == 'Your images should always have an alt tag, because it improves accessibility for visually impaired people. The name of the file is important too, because it helps Google understand what your image is about. For example, you could rename a file named "IMG0001.jpg" to "tree_with_a_bird.jpg".'
            )


def test_image_missing():
    from django_check_seo.checks_list import check_images

    site = init()
    check_images.run(site)

    for problem in site.problems:
        if problem.name == "Img lack alt tag":
            assert problem.name == "Img lack alt tag"
            assert problem.settings == "all images"
            assert problem.found == 1
            assert problem.searched_in == [
                '<b><u class="problem"><a target="_blank" href="none">image</a></u></b>'
            ]


def test_images_missing():
    import copy

    from django_check_seo.checks_list import check_images

    site = init()
    site.soup.find("img")["alt"] = ""
    site.soup.body.append(copy.copy(site.soup.find("img")))
    check_images.run(site)

    for problem in site.problems:
        if problem.name == "Img lack alt tag":
            assert problem.name == "Img lack alt tag"
            assert problem.settings == "all images"
            assert problem.found == 2
            assert problem.searched_in == [
                '<b><u class="problem"><a target="_blank" href="none">image</a></u></b>',
                '<b><u class="problem"><a target="_blank" href="none">image</a></u></b>',
            ]


def test_image_unknown():
    from django_check_seo.checks_list import check_images

    site = init()
    site.soup.find("img")["src"] = ""
    site.soup.find("img")["alt"] = ""
    check_images.run(site)

    for problem in site.problems:
        if problem.name == "Img lack alt tag":
            assert problem.name == "Img lack alt tag"
            assert problem.settings == "all images"
            assert problem.found == 1
            assert problem.searched_in == [
                "<b><u>unknown image</u></b>",
            ]



# see https://github.com/kapt-labs/django-check-seo/wiki/Toolbar-shortcut#cms_toolbarspy

# Third party
from cms.toolbar_base import CMSToolbar
from cms.toolbar_pool import toolbar_pool

try:
    from django.utils.translation import ugettext_lazy as _
except ImportError:
    from django.utils.translation import gettext_lazy as _


class DjangoSeoToolbar(CMSToolbar):
    def populate(self):
        # display only if user has permission to access the django-check-seo app
        if self.request.user.has_perm("django_check_seo.use_django_check_seo"):
            self.toolbar.add_sideframe_item(
                _("Check SEO"),
                "/django-check-seo/?page=" + self.request.path,
            )


# register the toolbar
toolbar_pool.register(DjangoSeoToolbar)



# Generated manually for django_cms_model_keywords

from django.db import migrations, models
from django.db.models import Q


class Migration(migrations.Migration):
    dependencies = [
        ("django_check_seo", "0002_keyword_page"),
    ]

    operations = [
        migrations.AlterField(
            model_name="page",
            name="path",
            field=models.CharField(
                blank=True,
                db_index=True,
                max_length=500,
                null=True,
                unique=False,
                verbose_name="path",
            ),
        ),
        migrations.AddField(
            model_name="page",
            name="cms_page_id",
            field=models.PositiveIntegerField(
                blank=True,
                db_index=True,
                null=True,
                verbose_name="CMS page id",
            ),
        ),
        migrations.AddField(
            model_name="page",
            name="language",
            field=models.CharField(
                blank=True, max_length=10, null=True, verbose_name="language"
            ),
        ),
        migrations.AddConstraint(
            model_name="page",
            constraint=models.UniqueConstraint(
                condition=Q(path__isnull=False) & ~Q(path=""),
                fields=("path",),
                name="dcs_unique_page_path_when_set",
            ),
        ),
        migrations.AddConstraint(
            model_name="page",
            constraint=models.UniqueConstraint(
                condition=Q(cms_page_id__isnull=False),
                fields=("cms_page_id", "language"),
                name="dcs_unique_cms_page_language",
            ),
        ),
    ]



# Tests for model_keywords and django_cms_model_keywords discovery (require DB)

import pytest
from bs4 import BeautifulSoup

from django_check_seo.models import Keyword, Page
from django_check_seo.utils.keywords_discovery import (
    django_cms_model_keywords,
    model_keywords,
)

html_content = """
<!doctype html>
<html><head></head><body></body></html>
"""


def _make_site(path="/fr/ma-page/"):
    obj = type("Site", (), {})()
    obj.keywords = []
    obj.soup = BeautifulSoup(html_content, features="lxml")
    obj.full_url = "https://example.com" + path
    return obj


@pytest.mark.django_db
def test_model_keywords_found():
    """model_keywords sets site.keywords from Page/Keyword for matching path."""
    kw1 = Keyword.objects.create(name="foo")
    kw2 = Keyword.objects.create(name="bar")
    p = Page.objects.create(path="/fr/ma-page/")
    p.keywords.add(kw1, kw2)

    site_obj = _make_site("/fr/ma-page/")
    model_keywords(site_obj)
    assert site_obj.keywords == ["foo", "bar"]


@pytest.mark.django_db
def test_model_keywords_unknown_path():
    """model_keywords leaves site.keywords empty when no Page for path."""
    site_obj = _make_site("/unknown/path/")
    model_keywords(site_obj)
    assert site_obj.keywords == []


@pytest.mark.django_db
def test_model_keywords_empty_page():
    """model_keywords sets site.keywords to [] when Page has no keywords."""
    Page.objects.create(path="/fr/empty/")
    site_obj = _make_site("/fr/empty/")
    model_keywords(site_obj)
    assert site_obj.keywords == []


@pytest.mark.django_db
def test_model_keywords_path_with_leading_slash():
    """Path from URL is used as-is when it starts with /."""
    kw = Keyword.objects.create(name="baz")
    p = Page.objects.create(path="/no-leading-slash")
    p.keywords.add(kw)

    site_obj = _make_site("/no-leading-slash")
    model_keywords(site_obj)
    assert site_obj.keywords == ["baz"]


# django_cms_model_keywords: without CMS, falls back to path (same as model_keywords)
@pytest.mark.django_db
def test_django_cms_model_keywords_fallback_path():
    """django_cms_model_keywords uses path when path is not a CMS page (e.g. apphook)."""
    kw1 = Keyword.objects.create(name="apphook-kw")
    p = Page.objects.create(path="/my-apphook/")
    p.keywords.add(kw1)
    site_obj = _make_site("/my-apphook/")
    django_cms_model_keywords(site_obj)
    assert site_obj.keywords == ["apphook-kw"]


@pytest.mark.django_db
def test_django_cms_model_keywords_by_cms_id():
    """django_cms_model_keywords uses cms_page_id+language when page has them."""
    kw1 = Keyword.objects.create(name="cms-kw")
    p = Page.objects.create(cms_page_id=42, language="fr", path=None)
    p.keywords.add(kw1)
    # Without CMS in the project, get_cms_page_id_language_for_path returns (None, None),
    # so we cannot test the CMS resolution here. Just check that a Page with only
    # cms_page_id+language has keywords and get_display_identifier works.
    assert [kw.name for kw in p.keywords.all()] == ["cms-kw"]
    assert "42" in p.get_display_identifier() and "fr" in p.get_display_identifier()



# Property-based (fuzzy) tests for keyword-related logic.

import pytest

try:
    from hypothesis import given
    from hypothesis import strategies as st
except ImportError:
    pytest.skip("hypothesis not installed", allow_module_level=True)

# Keywords and text are used in regex patterns; avoid regex-special characters.
_safe_text = st.text(
    alphabet=st.characters(
        whitelist_categories=("L", "N", "Zs"),
        whitelist_characters="'-",
    ),
    max_size=300,
)
_safe_keyword = st.text(
    alphabet=st.characters(
        whitelist_categories=("L", "N"),
        whitelist_characters="'-",
    ),
    min_size=1,
    max_size=30,
)


@given(
    text=st.text(alphabet=st.characters(blacklist_categories=("Cs",)), max_size=2000)
)
def test_parse_keywords_text_returns_list_of_non_empty_strings(text):
    """_parse_keywords_text always returns a list of non-empty strings."""
    from django_check_seo.views import _parse_keywords_text

    result = _parse_keywords_text(text)
    assert isinstance(result, list)
    for item in result:
        assert isinstance(item, str)
        assert len(item) > 0
        assert item == item.strip()


@given(text=st.one_of(st.none(), st.just(""), st.just("   "), st.just("\n\t")))
def test_parse_keywords_text_empty_or_whitespace_returns_empty_list(text):
    """_parse_keywords_text returns [] for None, empty or whitespace-only string."""
    from django_check_seo.views import _parse_keywords_text

    assert _parse_keywords_text(text) == []


@given(
    keywords=st.lists(
        st.text(
            alphabet=st.characters(blacklist_categories=("Cs",)),
            min_size=1,
            max_size=50,
        ),
        min_size=0,
        max_size=20,
    )
)
def test_parse_keywords_text_roundtrip_via_comma(keywords):
    """Parsing a comma-joined string of keywords yields the same list (modulo order if no dupes)."""
    from django_check_seo.views import _parse_keywords_text

    if not keywords:
        return
    # Avoid commas/newlines in keywords for a simple roundtrip
    safe = [k.replace(",", " ").replace("\n", " ").strip() for k in keywords]
    safe = [k for k in safe if k]
    if not safe:
        return
    text = ", ".join(safe)
    result = _parse_keywords_text(text)
    assert sorted(result) == sorted(safe)


@given(s=st.text(alphabet=st.characters(blacklist_categories=("Cs",)), max_size=500))
def test_normalize_apostrophes_no_crash(s):
    """normalize_apostrophes accepts any string without raising."""
    from django_check_seo.checks.utils import normalize_apostrophes

    out = normalize_apostrophes(s)
    assert isinstance(out, str)
    assert "'" not in out or "\u2019" in out


@given(
    text=_safe_text,
    keywords=st.lists(_safe_keyword, min_size=0, max_size=15),
)
def test_highlight_keywords_in_text_shape(text, keywords):
    """highlight_keywords_in_text returns (str, list of ints, str); occurrence list length = len(keywords)."""
    from django_check_seo.checks.utils import highlight_keywords_in_text

    highlighted, occurrence_list, found_str = highlight_keywords_in_text(
        text.lower(), keywords, normalize_apostrophes_flag=False
    )
    assert isinstance(highlighted, str)
    assert isinstance(occurrence_list, list)
    assert len(occurrence_list) == len(keywords)
    for n in occurrence_list:
        assert isinstance(n, int)
        assert n >= 0
    assert isinstance(found_str, str)


@given(keyword=_safe_keyword, text=_safe_text)
def test_count_keyword_occurrences_non_negative(keyword, text):
    """count_keyword_occurrences returns a non-negative integer."""
    from django_check_seo.checks.utils import count_keyword_occurrences

    n = count_keyword_occurrences(keyword.lower(), text.lower(), for_url=False)
    assert isinstance(n, int)
    assert n >= 0



from django.contrib import admin

from .models import Keyword, Page


@admin.register(Page)
class PageAdmin(admin.ModelAdmin):
    list_display = ("path", "nb_keywords")


@admin.register(Keyword)
class KeywordAdmin(admin.ModelAdmin):
    list_display = ("name", "pages_using")
